# -*- coding: utf-8 -*-
"""
 BASEmesh: a QGIS plugin for the creation of computational meshes
 for the numerical modelling software BASEMENT using Jonathan Shewchuk's mesh
 generator 'Triangle'.

 Copyright (C) 2013 Florian Hinkelammert and Christian Volz, ETH Zurich
 hinkelammert@vaw.baug.ethz.ch / volz@vaw.baug.ethz.ch

 This QGIS Plugin is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 2
 of the License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

 For the mesh generator Triangle, the licence regulations are not obvious. It's
 author Jonathan Shevchuk has the copyright on the software and want's to be
 informed if the software is used. For more information, visit
 http://www.cs.cmu.edu/~quake/triangle.html.
"""    

from qgis.core import *
from qgis.utils import *
from PyQt4.QtCore import *
from PyQt4.QtGui import *

import math
import numpy as np
import random as rnd

from ..tools.mesh import MESH
from ..tools import commonFunctions
from ..pyqtgl import opengl


class MESHCONVERSION:
    
    def __init__ (self):
        pass
    
    def getMesh (self):
        return self.__mesh
    
    def setMesh (self, mesh):
        self.__mesh = mesh
        pass

    def readTriangle (self, fileID, progressBar = None):
        self.__mesh = MESH()
        # import of fileID.node and fileID.ele files, generated by TRIANGLE
        # parse node file
        try:
            nodeFile = open(fileID+'.node', 'r')
        except:
            return  ('could not open the file ''%s.node''!' % fileID)
        line = nodeFile.readline().split()
        N = int(float(line[0]))
        if int(float(line[2])) == 1:
            for ii in range(0,N):
                line = [float(l) for l in nodeFile.readline().split()]
                self.__mesh.addNode( int(line[0]+1), line[1], line[2], line[3] ) # increase node ID by 1, because TRIANGLE starts at 0, here we need 1 (SMS convention)
        else:
            for ii in range(0,N):
                line = [float(l) for l in nodeFile.readline().split()]
                self.__mesh.addNode( int(line[0]+1), line[1], line[2] ) # increase node ID by 1, because TRIANGLE starts at 0, here we need 1 (SMS convention)
        # parse element file
        self.__mesh.newAttribute('matid')
        try:
            eleFile = open(fileID+'.ele', 'r')
        except:
            return ('could not open the file ''%s.ele''!' % fileID)
        line = eleFile.readline().split()
        N = int(float(line[0]))
        if progressBar != None:
            progressBar.setRange(0, N)
            progressBar.setValue(0)
        for ii in range(0,N):
            line = [int(float(l)) for l in eleFile.readline().split()]
            elID = line[0]+1 # increase element ID by 1, because TRIANGLE starts at 0, here we need 1 (SMS convention)
            nodeIDs = [line[1]+1,line[2]+1,line[3]+1] # increase node ID by 1 (see above when adding a NODE)
            self.__mesh.addElement( elID, nodeIDs )
            if len(line) == 5:
                self.__mesh.setAttr(elID, ('matid', line[4]))
            else:
                self.__mesh.setAttr(elID, ('matid', 0))
            if progressBar!=None:
                commonFunctions.updateProgressBar(progressBar)
        # return success
        success = self.__mesh.init(1)
        if success!=True:
            return success
        else:
            return True

    def read2dm (self,  fileID):
        self.__mesh = MESH()
        # import .2dm file, generated by SMS or QGIS
        try:
            smsFile = open(fileID+'.2dm', 'r')
        except:
            return ("could not open the file '%s.2dm'!" % fileID)
        lineCounter = 0
        elData = []
        for line in smsFile:
            # parse line
            tokens = line.strip().split()
            if len(tokens) < 1:
                continue
            if tokens[0] in ["MESH2d", "MESH2D", "MESHNAME"]:
                continue
            # store triangle data in list
            if tokens[0] == "E3T":
                if len(tokens) < 6:
                    return ("Error found in 2dm file at cell-definition at line #%d. Cannot proceed." % lineCounter+1)
                else:
                    elData.append( [int(tokens[1]), int(tokens[5]), int(tokens[2]), int(tokens[3]), int(tokens[4])] )
            # store quadrilateral data in list
            elif tokens[0] == "E4Q":
                if len(tokens) < 7:
                    return ("Error found in 2dm file at cell-definition at line #%d. Cannot proceed." % lineCounter+1)
                else:
                    elData.append( [int(tokens[1]), int(tokens[6]), int(tokens[2]), int(tokens[3]), int(tokens[4]), int(tokens[5])] )
            # add node to mesh
            elif tokens[0] == "ND":
                if len(tokens) < 5:
                    return ("Error found in 2dm file at node-definition at line #%d. Cannot proceed." % lineCounter+1)
                else:
                    self.__mesh.addNode( int(tokens[1]), float(tokens[2]), float(tokens[3]), float(tokens[4]) )
            # go to next line
            lineCounter = lineCounter + 1
        smsFile.close()
        # add all cells to the mesh
        self.__mesh.newAttribute('matid')
        for eli in elData:
            self.__mesh.addElement( eli[0], eli[2::] )
            # and define material id
            self.__mesh.setAttr( eli[0], ('matid', eli[1]) )
        # return success
        #return "llll"
        success = self.__mesh.init()
        if success!=True:
            return success
        else:
            return True

    def readShape (self, nodeLayer, elevationField, meshLayer, materialField, progressBar=None):
        self.__mesh = MESH()
        # create mesh object from active layers in QGIS
        # parse point layer
        # do some work before the main loop
        if nodeLayer.geometryType()==QGis.Point:
            nodeIDindex = nodeLayer.fieldNameIndex('NODE_ID')
            elevationIndex = nodeLayer.fieldNameIndex(elevationField)
        else:
            return ("layer '%s' not of type 'Point'!"%nodeLayer.name())
        # retreive every feature with its geometry and attributes
        for feat in nodeLayer.getFeatures():
            point = feat.geometry().asPoint()
            attrs = feat.attributes()
            nid = attrs[nodeIDindex]
            z = attrs[elevationIndex]
            self.__mesh.addNode( nid, point[0], point[1], z )
            # update progress bar
            if progressBar!=None:
                commonFunctions.updateProgressBar(progressBar)
        # parse polygon layer
        # do some work before the main loop
        self.__mesh.newAttribute('matid')
        if meshLayer.geometryType() == QGis.Polygon:
            elementIDindex = meshLayer.fieldNameIndex('ELEMENT_ID')
            if materialField != None:
                materialIndex = meshLayer.fieldNameIndex(materialField)
            else:
                materialIndex = -1
            nodeIDindices = []
            for ii in range(1, 5):
                nodeIDindices.append( meshLayer.fieldNameIndex('NODE%i'%ii) )
        else:
            return ("layer '%s' not of type 'Polygon'!"%meshLayer.name())
        # do all the work
        for feat in meshLayer.getFeatures():
            attrs = feat.attributes()
            elid = attrs[elementIDindex]
            # get node ids
            nodeIDs = []
            for ii in range(0, 4):
                nid = attrs[nodeIDindices[ii]]
                if nid>0:
                    nodeIDs.append(nid)
            # add cell to the mesh
            self.__mesh.addElement( elid, nodeIDs )
            # and eventually define material id
            if materialIndex >= 0:
                self.__mesh.setAttr( elid, ('matid', attrs[materialIndex]) )
            # update progress bar
            if progressBar!=None:
                commonFunctions.updateProgressBar(progressBar)
        # return success
        success = self.__mesh.init()
        if success!=True:
            return success
        else:
            return True

    def write2dm (self, outputFileID, progressBar=None):
        # writes the mesh into a .2dm-file
        file2dm = open(outputFileID+'.2dm',  "w")
        try:
            file2dm.write("MESH2D   #created automatically via BASEmesh-Plugin for QGIS\n")
            # write all elements
            for elID in self.__mesh.getElementIDs():
                nodeIDs = self.__mesh.getElementNodeIDs(elID)
                material = self.__mesh.getAttr(elID, 'matid')
                if len(nodeIDs) == 3:
                    file2dm.write( "E3T %7d %7d %7d %7d %7d\n" % (elID, nodeIDs[0], nodeIDs[1], nodeIDs[2], material) )
                elif len(nodeIDs) == 4:
                    file2dm.write("E4Q %7d %7d %7d %7d %7d %7d\n" % (elID, nodeIDs[0], nodeIDs[1], nodeIDs[2], nodeIDs[3], material) )
                else:
                    return ("conversion failed in element %i"%elID)
                if progressBar!=None:
                    commonFunctions.updateProgressBar(progressBar)
            # write all nodes
            for nID in self.__mesh.getNodeIDs():
                node = self.__mesh.getNode(nID)
                file2dm.write( "ND %7d %9f %9f %9f\n" % (nID, node.getX(), node.getY(), node.getZ()) )
                if progressBar!=None:
                    commonFunctions.updateProgressBar(progressBar)
        finally:
            file2dm.close()
        if not os.path.exists(outputFileID+'.2dm'):
            return ("2dm-file could not be created.")
        else:
            return True
    
    def writeShape (self, outputFileID,  progressBar=None):
        # get coordinate reference system of the project
        crs = iface.mapCanvas().mapRenderer().destinationCrs()
        # create the new shape file for nodes
        fields = QgsFields()
        fields.append(QgsField("NODE_ID", QVariant.Int))
        fields.append(QgsField("X", QVariant.Double))
        fields.append(QgsField("Y", QVariant.Double))
        fields.append(QgsField("Z", QVariant.Double))
        try:
            writer = QgsVectorFileWriter(outputFileID+'_nodes', "CP1250", fields, QGis.WKBPoint, crs, "ESRI Shapefile")
            for ptID in self.__mesh.getNodeIDs():
                ptFeat = QgsFeature(fields) # create destination feature
                ptFeat.setFeatureId(ptID)
                # define feature geometry
                pt = self.__mesh.getNode(ptID)
                ptFeat.setGeometry(QgsGeometry.fromPoint( QgsPoint( pt.getX(), pt.getY() ) ))
                # fill fields with contents of nodefile
                ptFeat.setAttribute(0, ptID)
                ptFeat.setAttribute(1, pt.getX())
                ptFeat.setAttribute(2, pt.getY())
                ptFeat.setAttribute(3, pt.getZ())
                writer.addFeature(ptFeat)
                if progressBar!=None:
                    commonFunctions.updateProgressBar(progressBar)
        finally:
            # delete the writer to flush features to disk (optional)
            del writer
        if not os.path.exists(outputFileID+'_nodes.shp'):
            return ("shape-file with nodes could not be created.")
        # create the new shape file for cells
        fields = QgsFields()
        fields.append(QgsField("ELEMENT_ID", QVariant.Int))
        fields.append(QgsField("NODE1", QVariant.Int))
        fields.append(QgsField("NODE2", QVariant.Int))
        fields.append(QgsField("NODE3", QVariant.Int))
        fields.append(QgsField("NODE4", QVariant.Int))
        fields.append(QgsField("MATID", QVariant.Int))
        try:
            writer = QgsVectorFileWriter(outputFileID+'_elements', "CP1250", fields, QGis.WKBPolygon, crs, "ESRI Shapefile")
            for elID in self.__mesh.getElementIDs():
                elFeat = QgsFeature(fields) # create destination feature
                elFeat.setFeatureId(elID)
                elFeat.setAttribute(0, elID) # write ID to attribute field
                nodeIDs = self.__mesh.getElementNodeIDs(elID)
                qgsPointList = []
                for ii in range(0, len(nodeIDs)):
                    elFeat.setAttribute(ii+1, nodeIDs[ii]) # write node IDs to attribute field
                    pt = self.__mesh.getNode(nodeIDs[ii])
                    qgsPointList.append( QgsPoint(pt.getX(), pt.getY()) )
                # in case of triangular element
                if len(nodeIDs)<4:
                    elFeat.setAttribute(4, -1)
                # write material ID to attribute field
                elFeat.setAttribute( 5, self.__mesh.getAttr(elID, 'matid') )
                # define feature geometry
                elFeat.setGeometry(QgsGeometry.fromPolygon( [ qgsPointList ] ))
                # finally add the feature
                writer.addFeature(elFeat)
                if progressBar!=None:
                    commonFunctions.updateProgressBar(progressBar)
        finally:
            # delete the writer to flush features to disk (optional)
            del writer
        if not os.path.exists(outputFileID+'_elements.shp'):
            return ("shape-file with cells could not be created.")
        else:
            return True

    def writeOpenGL (self,zScale=1.0):
        # test function: plot the mesh in a opengl window...yeah!!!
        # write all nodes
        verts = []
        nodeIDs = {}
        ii = 0
        for nID in self.__mesh.getNodeIDs():
            node = self.__mesh.getNode(nID)
            verts.append([node.getX(), node.getY(), zScale*node.getZ()])
            nodeIDs[nID] = ii
            ii = ii+1
        # write all elements
        faces = []
        faceColor = []
        foundMatids = [1]
        matColor = [[0.5,0.5,0.5,1]]
        for elID in self.__mesh.getElementIDs():
            nids = self.__mesh.getElementNodeIDs(elID)
            material = self.__mesh.getAttr(elID, 'matid')
            if len(nids) == 3:
                faces.append([nodeIDs[nids[0]], nodeIDs[nids[1]], nodeIDs[nids[2]]])
                if material not in foundMatids:
                    foundMatids.append(material)
                    matColor.append([rnd.random(),rnd.random(),rnd.random(),1])
                faceColor.append(matColor[foundMatids.index(material)])
            else:
                return ("conversion failed in element %i. only triangles are supported up to now."%elID)
        # plot that shit
        verts = np.array(verts)
        dx = 0.5*(verts[:,0].max()+verts[:,0].min())
        dy = 0.5*(verts[:,1].max()+verts[:,1].min())
        dz = 0.5*(verts[:,2].max()+verts[:,2].min())
        view = opengl.GLViewWidget()
        view.setGeometry(400,250,1200,700)
        view.setWindowIcon(QIcon(os.path.join(commonFunctions.getPluginDir(),"icons/showmesh.svg")))
        view.setWindowTitle('3D mesh preview')
        view.setBackgroundColor([1,1,1])
        mesh = opengl.MeshData(vertexes=verts,faces=np.array(faces),faceColors=np.array(faceColor))
        meshItem = opengl.GLMeshItem(meshdata=mesh,drawEdges=True,edgeWidth=3,edgeColor=(0.1,0.1,0.1,1),smooth=False)
        view.addItem(meshItem)
        view.pan(dx,dy,dz)
        view.show()
        return True